/******************************************************************************
    Serial port smart monitor for use with powerd-2.0 package or stand-alone.
    Can be set to read any of the status bits and signal
    OK/FAIL on any value or log in a file any message.

    This program is released under the GNU General Public License.
    See the file COPYING for details.
            
    I1EPJ 020815 - Rev. 040708 (Merge of {lp|ser|smart}monitor)
                   Rev. 040712 (Added support for parallel port A/D interface) 
		   Rev  041118 (Avoid inclusion of linux/lp.h for few #define) 
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <asm/ioctls.h>
#include <sys/io.h>
#include <sys/select.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

/*
 * Taken from linux/lp.h to not require having kernel headers installed
 * only for few #define
 */
#define LP_PBUSY        0x80  /* inverted input, active high */                 
#define LP_PACK         0x40  /* unchanged input, active low */                 
#define LP_POUTPA       0x20  /* unchanged input, active high */                
#define LP_PSELECD      0x10  /* unchanged input, active high */                
#define LP_PERRORP      0x08  /* unchanged input, active low */                 
#define LPGETSTATUS   0x060b  /* return LP_S(minor) */                            

/*
 * Configuration file
 */
#define CONFIG_FILE "/etc/smartmonitor.conf"
/*
 * FIFOs
 */
#define DIN1_FIFO "/var/spool/smartmonitor/din1"
#define DIN2_FIFO "/var/spool/smartmonitor/din2"
#define DIN3_FIFO "/var/spool/smartmonitor/din3"
#define DIN4_FIFO "/var/spool/smartmonitor/din4"
#define AIN1_FIFO "/var/spool/smartmonitor/ain1"
#define AIN2_FIFO "/var/spool/smartmonitor/ain2"
#define TEMP_FIFO "/var/spool/smartmonitor/temp"
#define VALIM_FIFO "/var/spool/smartmonitor/valim"

#define MSGLEN 23
#define NONE 0

const unsigned char O_SOC = 0x08;
unsigned char O_NLO = 0x00;
unsigned char O_NHI = 0x10;
unsigned char I_EOC = 0x80;
unsigned char I_MSK = 0x78;

typedef struct
{
	char *name;
	int value;
} bitname;

typedef struct
{
	char *token;
	int value;
} keyword;

bitname lp_bitnames[] = {
	{"PBUSY", LP_PBUSY},
	{"PACK", LP_PACK},
	{"POUTPA", LP_POUTPA},
	{"PSELECD", LP_PSELECD},
	{"PERRORP", LP_PERRORP},
	{"NONE", NONE},
	{NULL, 0}
};

bitname ser_bitnames[] = {
	{"CTS", TIOCM_CTS},
	{"DSR", TIOCM_DSR},
	{"DCD", TIOCM_CAR},
	{"RI", TIOCM_RNG},
	{"NONE", NONE},
	{NULL, 0}
};
bitname smart_bitnames[] = {
	{"DIN1", 0x01},
	{"DIN2", 0x02},
	{"DIN3", 0x04},
	{"DIN4", 0x08},
	{"NONE", NONE},
	{NULL, 0}
};

enum
{
	MONITOR_TYPE, DEVICE, LOGFILE,
	MAILTOP, MAILTO1, MAILTO2, MAILTO3, MAILTO4,
	POWERB, POWEROK, SPEED, SETOUT,
	AIN1MULT, AIN2MULT, TEMPMULT, VCCMULT, LOGTIME, UPDATETIME,
	ALARM1B, ALARM2B, ALARM3B, ALARM4B,
	ALARM1O, ALARM2O, ALARM3O, ALARM4O,
	POWERONT, POWEROFFT,
	ALARM1T0, ALARM2T0, ALARM3T0, ALARM4T0,
	ALARM1T1, ALARM2T1, ALARM3T1, ALARM4T1,
	ALARM1A0, ALARM2A0, ALARM3A0, ALARM4A0,
	ALARM1A1, ALARM2A1, ALARM3A1, ALARM4A1,
	LOG_STARTUP_FMT, LOG_DATE_FMT, LOG_EVENT_FMT, LOG_ACTION_FMT, LOG_ANALOG_VALUES_FMT,
	MAIL_DATE_FMT, MAIL_EVENT_FMT, MAIL_ACTION_FMT
};

enum
{
	PARALLEL, SERIAL, SMART_PARALLEL, SMART_SERIAL
};

keyword keywords[] = {
	{"monitor_type", MONITOR_TYPE},
	{"device", DEVICE},
	{"logfile", LOGFILE},
	{"mailto_power", MAILTOP},
	{"mailto_alarm1", MAILTO1},
	{"mailto_alarm2", MAILTO2},
	{"mailto_alarm3", MAILTO3},
	{"mailto_alarm4", MAILTO4},
	{"power_bit", POWERB},
	{"power_ok", POWEROK},
	{"set_out", SETOUT},
	{"speed", SPEED},
	{"ain1_mult", AIN1MULT},
	{"ain2_mult", AIN2MULT},
	{"temp_mult", TEMPMULT},
	{"vcc_mult", VCCMULT},
	{"log_time", LOGTIME},
	{"update_time", UPDATETIME},
	{"alarm1_bit", ALARM1B},
	{"alarm2_bit", ALARM2B},
	{"alarm3_bit", ALARM3B},
	{"alarm4_bit", ALARM4B},
	{"alarm1_on", ALARM1O},
	{"alarm2_on", ALARM2O},
	{"alarm3_on", ALARM3O},
	{"alarm4_on", ALARM4O},
	{"power_on_text", POWERONT},
	{"power_off_text", POWEROFFT},
	{"alarm1_on_text", ALARM1T1},
	{"alarm2_on_text", ALARM2T1},
	{"alarm3_on_text", ALARM3T1},
	{"alarm4_on_text", ALARM4T1},
	{"alarm1_off_text", ALARM1T0},
	{"alarm2_off_text", ALARM2T0},
	{"alarm3_off_text", ALARM3T0},
	{"alarm4_off_text", ALARM4T0},
	{"alarm1_on_action", ALARM1A1},
	{"alarm2_on_action", ALARM2A1},
	{"alarm3_on_action", ALARM3A1},
	{"alarm4_on_action", ALARM4A1},
	{"alarm1_off_action", ALARM1A0},
	{"alarm2_off_action", ALARM2A0},
	{"alarm3_off_action", ALARM3A0},
	{"alarm4_off_action", ALARM4A0},
	{"log_start_fmt_string", LOG_STARTUP_FMT},
	{"log_date_fmt_string", LOG_DATE_FMT},
	{"log_event_fmt_string", LOG_EVENT_FMT},
	{"log_action_fmt_string", LOG_ACTION_FMT},
	{"log_analog_values_fmt_string", LOG_ANALOG_VALUES_FMT},
	{"mail_date_fmt_string", MAIL_DATE_FMT},
	{"mail_event_fmt_string", MAIL_EVENT_FMT},
	{"mail_action_fmt_string", MAIL_ACTION_FMT},
	{NULL, 0}
};

/*
 * Global variables
 */
int type = PARALLEL;
char *smartmonitor_start_fmt;
char *log_date_fmt;
char *log_event_fmt;
char *log_action_fmt;
char *log_analog_values_fmt;
char *mail_date_fmt;
char *mail_event_fmt;
char *mail_action_fmt;

/*
 * Parse a string for C escape sequences. Taken from xNOS.
 */
char *stringparse(char *line)
{
	char *ret = line, *cp = line;
	unsigned long num;

	if (*line == '\"')
		line++;

	while (*line != '\0' && *line != '\"') {
		if (*line == '\\') {
			line++;
			switch (*line++) {
				case 'n':
					*cp++ = '\n';
					break;
				case 't':
					*cp++ = '\t';
					break;
				case 'v':
					*cp++ = '\v';
					break;
				case 'b':
					*cp++ = '\b';
					break;
				case 'r':
					*cp++ = '\r';
					break;
				case 'f':
					*cp++ = '\f';
					break;
				case 'a':
					*cp++ = '\a';
					break;
				case '\\':
					*cp++ = '\\';
					break;
				case '\?':
					*cp++ = '\?';
					break;
				case '\'':
					*cp++ = '\'';
					break;
				case '\"':
					*cp++ = '\"';
					break;
				case 'x':
					num = strtoul(line, &line, 16);
					*cp++ = (char) num;
					break;
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
					line = line - 1;
					num = strtoul(line, &line, 8);
					*cp++ = (char) num;
					break;
				case '\0':
					return NULL;
				default:
					*cp++ = *(line - 1);
					break;
			};
		} else {
			*cp++ = *line++;
		}
	}

	if (*line == '\"')
		line++;		/* skip final quote */
	*cp = '\0';		/* terminate string */
	return ret;
}

/*
 * Return bitmask value given bit name
 */
int bitvalue(char *bname)
{
	bitname *bp, *bp1;

	switch (type) {
		case PARALLEL:
			bp1 = lp_bitnames;
		case SERIAL:
			bp1 = ser_bitnames;
		case SMART_PARALLEL:
		case SMART_SERIAL:
			bp1 = smart_bitnames;
	}

	for (bp = bp1; bp->name != NULL; bp++) {
		if (strcasecmp(bp->name, bname) == 0) {
			return bp->value;
		}
	}
	fprintf(stderr, CONFIG_FILE ": Illegal bit name %s\n", bname);
	return -1;
}

/*
 * Write a message to log file
 */
void writelog(char *logfile, char *fmt, ...)
{
	va_list ap;
	FILE *fp;
	char *cp;
	time_t t;
	char buf[128];

	va_start(ap, fmt);

	if (logfile == NULL)
		return;

	if ((fp = fopen(logfile, "a+")) == NULL) {
		perror(logfile);
		exit(2);
	} else {
		time(&t);
		cp = ctime(&t);
		*(cp + strlen(cp) - 1) = '\0';
		vsprintf(buf, fmt, ap);
		fprintf(fp, log_date_fmt, cp);
		fprintf(fp, "%s\n", buf);
		fclose(fp);
	}
	va_end(ap);
}

/*
 * Notify sysop of an event
 */
void mailsysop(char *mailto, char *message, char *action)
{
	char buf[128];
	FILE *fp;
	char *cp;
	time_t t;

	if (mailto == NULL)
		return;

	/*
	 * Mail wanted, prepare and send it
	 */
	time(&t);
	cp = ctime(&t);
	*(cp + strlen(cp) - 1) = '\0';
	sprintf(buf, "mail -s smartmonitor-Report %s", mailto);
	if ((fp = popen(buf, "w")) != NULL) {
		fprintf(fp, mail_date_fmt, cp);
		fprintf(fp, mail_event_fmt, message);
		if (action != NULL) {
			fprintf(fp, mail_action_fmt, action);
		}
		pclose(fp);
	}
}


int read_lp_status(int devaddr)
{
	int status;

	status = inb(devaddr + 1);
	return status;

}

void write_lp_value(int devaddr, int out_value)
{
	outb(out_value, devaddr);
}

int ADConvert(int devaddr, int Channel)
{
	int status, lo, hi;

#ifdef DEBUG
	printf("Starting conversion on channel %d\n", Channel);
#endif
	/* Start a new conversion (pulse SOC) */
	write_lp_value(devaddr, O_NLO | Channel);
	write_lp_value(devaddr, O_NLO | O_SOC | Channel);
	write_lp_value(devaddr, O_NLO | Channel);

	/* Wait end of conversion */
	do {
		status = read_lp_status(devaddr);
#ifdef DEBUG
		printf("EOC=%d\n", status & LP_PBUSY);
#endif
	} while ((status & LP_PBUSY) == 0);

	/* Read value one nibble at a time */
	write_lp_value(devaddr, O_NLO | Channel);
	lo = (read_lp_status(devaddr) & I_MSK) >> 3;

	write_lp_value(devaddr, O_NHI | Channel);
	hi = (read_lp_status(devaddr) & I_MSK) >> 3;
#ifdef DEBUG
	printf("Values read: hi=%x lo=%x\n", hi, lo);
#endif
	return 16 * hi + lo;
}

int main(int argc, char *argv[])
{
	int lp, ser, status = 0, cmd, index, n, r, nfd, devaddr, logtime = 15, updatetime = 60, ntimes = 15;
	char *logfile = NULL, *mailto_power = NULL;
	char line[256], *cp1, *cp2;
	keyword *kp;
	FILE *fp;
	int fdin1, fdin2, fdin3, fdin4, fain1, fain2, ftemp, fvalim;
	int ain1 = 0, ain2 = 0, temp = 0, valim = 0;
	int a1, a2, t, v, s;
	float ain1_mult = 1.0, ain2_mult = 1.0, temp_mult = 50.0, vcc_mult = 15.0, value;
	char *device = NULL;
	int out_value = -1;
	int speed;
	int power_mask = 0;
	int power_ok = 0, power_fail = 0, power_status = 0;
	int alarm_mask[4] = { 0, 0, 0, 0 };
	int alarm_status[4] = { 0, 0, 0, 0 };
	int alarm_on[4] = { 0, 0, 0, 0 };
	int alarm_off[4] = { 0, 0, 0, 0 };
	char *power_on_text = NULL, *power_off_text = NULL;
	char *alarm_on_text[4] = { 0, 0, 0, 0 };
	char *alarm_off_text[4] = { 0, 0, 0, 0 };
	char *alarm_on_action[4] = { 0, 0, 0, 0 };
	char *alarm_off_action[4] = { 0, 0, 0, 0 };
	char *mailto_alarm[4] = { 0, 0, 0, 0 };
	fd_set readfd, writefd;
	struct termios tio;
	char buf[256];
	struct timeval tv;

	/*
	 * Read config file
	 */
	if ((fp = fopen(CONFIG_FILE, "r")) == NULL) {
		perror(CONFIG_FILE);
		exit(2);
	} else {
		while (fgets(line, 256, fp), !feof(fp)) {
#ifdef DEBUG
			printf("Config line read: %s", line);
#endif
			if (line[strlen(line) - 1] == '\n')
				line[strlen(line) - 1] = '\0';

			if (*line == '#' || *line == '\0')
				continue;

			cp1 = strtok(line, "=");
			for (cmd = -1, kp = keywords; kp->token != NULL; kp++) {
				if (strcasecmp(kp->token, cp1) == 0) {
					cmd = kp->value;
					break;
				}
			}

			cp2 = strtok(NULL, "\n");
			if (strcasecmp(cp2, "NONE") == 0)
				continue;

			switch (cmd) {
				case MONITOR_TYPE:
					if (strcasecmp(cp2, "parallel") == 0)
						type = PARALLEL;
					else if (strcasecmp(cp2, "serial") == 0)
						type = SERIAL;
					else if (strcasecmp(cp2, "smart_parallel") == 0)
						type = SMART_PARALLEL;
					else if (strcasecmp(cp2, "smart_serial") == 0)
						type = SMART_SERIAL;
					else
						fprintf(stderr, "smartmonitor.conf: unknown hardware type %s\n", cp2);
				case DEVICE:
					device = strdup(cp2);
					if (strncasecmp(device, "0x", 2) == 0) {
						devaddr = strtoul(cp2, NULL, 16);
					}
					break;
				case LOGFILE:
					logfile = strdup(cp2);
					break;
				case MAILTOP:
					mailto_power = strdup(cp2);
					break;
				case MAILTO1:
				case MAILTO2:
				case MAILTO3:
				case MAILTO4:
					mailto_alarm[cmd - MAILTO1] = strdup(cp2);
					break;
				case POWERB:
					power_mask = bitvalue(cp2);
					break;
				case POWEROK:
					if (*cp2 == '1') {
						power_ok = power_mask;
						power_fail = 0;
					} else {
						power_ok = 0;
						power_fail = power_mask;
					}
					power_status = power_ok;
					break;
				case SPEED:
					speed = atoi(cp2);
					break;
				case SETOUT:
					if (type == PARALLEL) {
						out_value = strtoul(cp2, NULL, 16);
					}
					if (type == SERIAL) {
						if (*cp2 == 'H') {
							out_value = 0;
						} else if (*cp2 == 'L') {
							out_value = TIOCM_RTS | TIOCM_DTR;
						} else {
							out_value = -1;
						}
					}
					break;
				case AIN1MULT:
					ain1_mult = atof(cp2);
					break;
				case AIN2MULT:
					ain2_mult = atof(cp2);
					break;
				case TEMPMULT:
					temp_mult = atof(cp2);
					break;
				case VCCMULT:
					vcc_mult = atof(cp2);
					break;
				case UPDATETIME:
					updatetime = atoi(cp2);
					break;
				case LOGTIME:
					ntimes = logtime = atoi(cp2);
					break;
				case ALARM1B:
				case ALARM2B:
				case ALARM3B:
				case ALARM4B:
					alarm_mask[cmd - ALARM1B] = bitvalue(cp2);
					break;
				case ALARM1O:
				case ALARM2O:
				case ALARM3O:
				case ALARM4O:
					index = cmd - ALARM1O;
					if (*cp2 == '1') {
						alarm_on[index] = alarm_mask[index];
						alarm_off[index] = 0;
					} else {
						alarm_on[index] = 0;
						alarm_off[index] = alarm_mask[index];
					}
					alarm_status[index] = alarm_on[index];
					break;
				case POWERONT:
					power_on_text = strdup(stringparse(cp2));
					break;
				case POWEROFFT:
					power_off_text = strdup(stringparse(cp2));
					break;
				case ALARM1T1:
				case ALARM2T1:
				case ALARM3T1:
				case ALARM4T1:
					index = cmd - ALARM1T1;
					alarm_on_text[index] = strdup(stringparse(cp2));
					break;
				case ALARM1T0:
				case ALARM2T0:
				case ALARM3T0:
				case ALARM4T0:
					index = cmd - ALARM1T0;
					alarm_off_text[index] = strdup(stringparse(cp2));
					break;
				case ALARM1A1:
				case ALARM2A1:
				case ALARM3A1:
				case ALARM4A1:
					index = cmd - ALARM1A1;
					alarm_on_action[index] = strdup(cp2);
					break;
				case ALARM1A0:
				case ALARM2A0:
				case ALARM3A0:
				case ALARM4A0:
					index = cmd - ALARM1A0;
					alarm_off_action[index] = strdup(cp2);
					break;
				case LOG_STARTUP_FMT:
					smartmonitor_start_fmt = strdup(stringparse(cp2));
					break;
				case LOG_DATE_FMT:
					log_date_fmt = strdup(stringparse(cp2));
					break;
				case LOG_EVENT_FMT:
					log_event_fmt = strdup(stringparse(cp2));
					break;
				case LOG_ACTION_FMT:
					log_action_fmt = strdup(stringparse(cp2));
					break;
				case LOG_ANALOG_VALUES_FMT:
					log_analog_values_fmt = strdup(stringparse(cp2));
					break;
				case MAIL_DATE_FMT:
					mail_date_fmt = strdup(stringparse(cp2));
					break;
				case MAIL_EVENT_FMT:
					mail_event_fmt = strdup(stringparse(cp2));
					break;
				case MAIL_ACTION_FMT:
					mail_action_fmt = strdup(stringparse(cp2));
					break;
				default:
					fprintf(stderr, "smartmonitor.conf: illegal keyword %s\n", cp1);
					break;
			}
		}
	}

	writelog(logfile, smartmonitor_start_fmt, device);

#ifdef DEBUG
	char *sp;

	switch (type) {
		case PARALLEL:
			sp = "PARALLEL";
			break;
		case SERIAL:
			sp = "SERIAL";
			break;
		case SMART_PARALLEL:
			sp = "SMART_PARALLEL";
			break;
		case SMART_SERIAL:
			sp = "SMART_SERIAL";
			break;
	}
	printf("Type is %s\n", sp);
#endif

	if (type == SMART_PARALLEL) {
#ifdef DEBUG
		printf("Gaining access to port 0x%2.2x...\n", devaddr);
#endif
		if (ioperm(devaddr, 3, 1)) {
			perror("ioperm");
			exit(1);
		}

		n = 0;
		/* set all outputs to zero */
		write_lp_value(devaddr, 0);

		/* wait 1s for hardware to settle */
		sleep(1);
#ifdef DEBUG
		printf("Checking interface...\n");
#endif

		if ((read_lp_status(devaddr) & LP_PBUSY) == LP_PBUSY) {
			write_lp_value(devaddr, O_SOC);
			write_lp_value(devaddr, 0);
			if ((read_lp_status(devaddr) & LP_PBUSY) != 0) {
				printf("Error: hardware not responding properly...\n");
				exit(7);
			}
		}
	} else if (type == SMART_SERIAL) {
#ifdef DEBUG
		printf("Opening smart serial device %s...\n", device);
#endif
		if ((ser = open(device, O_RDWR)) == -1) {
			perror("open()");
			exit(3);
		}
		tcgetattr(ser, &tio);
		cfmakeraw(&tio);
		switch (speed) {
			case 1200:
				cfsetospeed(&tio, B1200);
				cfsetispeed(&tio, B1200);
				break;
			case 2400:
				cfsetospeed(&tio, B2400);
				cfsetispeed(&tio, B2400);
				break;
			case 4800:
				cfsetospeed(&tio, B4800);
				cfsetispeed(&tio, B4800);
				break;
			case 9600:
				cfsetospeed(&tio, B9600);
				cfsetispeed(&tio, B9600);
				break;
			default:
				cfsetospeed(&tio, B1200);
				cfsetispeed(&tio, B1200);
				break;
		}
		tcsetattr(ser, TCSANOW, &tio);
	} else if (type == PARALLEL) {
#ifdef DEBUG
		printf("Opening parallel device %s...\n", device);
#endif
		if ((lp = open(device, O_RDWR)) == -1) {
			perror("open()");
			exit(3);
		}

		if (out_value >= 0) {

			/* Check for hardware, if not ready in 1000 tries, give up */
			n = 0;
			do {
				status = read_lp_status(lp);
				n = n + 1;
			} while ((n < 1000) && ((status & LP_PBUSY) == 0));

			if (n == 1000) {
				printf("out_value != 0 but timeout waiting for LP_PBUSY...\nCheck if hardware is connected and running properly.\n");
				exit(6);
			}
#ifdef DEBUG
			printf("Writing 0x%2.2x to lp device...\n", out_value);
#endif
			if (write(lp, &out_value, 1) != 1) {
				perror("write()");
				exit(4);
			}
		}

	} else if (type == SERIAL) {
#ifdef DEBUG
		printf("Opening serial device %s...\n", device);
#endif
		if ((ser = open(device, O_RDWR)) == -1) {
			perror("open()");
			exit(3);
		}

		if (out_value != -1) {
#ifdef DEBUG
			printf("Setting %s CTS & DTR bits...\n", out_value ? "LOW" : "HIGH");
#endif
			if (ioctl(ser, TIOCMSET, &out_value) != 0) {
				perror("ioctl()");
				exit(4);
			}
		}
	} else {
		fprintf(stderr, "Unknown monitor type %d\n", type);
	}
	/*
	 * Open FIFOs
	 */
#ifdef DEBUG
	printf("Opening FIFOs...\n");
#endif
	if ((fdin1 = open(DIN1_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((fdin2 = open(DIN2_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((fdin3 = open(DIN3_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((fdin4 = open(DIN4_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((fain1 = open(AIN1_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((fain2 = open(AIN2_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((ftemp = open(TEMP_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
	if ((fvalim = open(VALIM_FIFO, O_RDWR)) == -1) {
		perror("open()");
		exit(4);
	}
#ifdef DEBUG
	printf("Entering for loop...\n");
#endif
	for (;;) {
		/* 
		 * Prepare for the select() call
		 */
		FD_ZERO(&readfd);
		FD_ZERO(&writefd);
		FD_SET(fdin1, &writefd);
		FD_SET(fdin2, &writefd);
		FD_SET(fdin3, &writefd);
		FD_SET(fdin4, &writefd);
		FD_SET(fain1, &writefd);
		FD_SET(fain2, &writefd);
		FD_SET(ftemp, &writefd);
		FD_SET(fvalim, &writefd);
		if (type == SMART_SERIAL) {
			FD_SET(ser, &readfd);
			/*
			 * Call select()
			 * Program waits here for a request from clients
			 * or data from serial smart interface
			 */
#ifdef DEBUG
			printf("Calling select()...?\n");
#endif
			nfd = select(ser + 1, &readfd, &writefd, NULL, NULL);
		} else {
			/*
			 * For all dumb interfaces and the smart parallel A/D interface
			 * define a 1s timeout to go to read data from it.
			 */
#ifdef DEBUG
			printf("Calling select()...\n");
#endif
			tv.tv_sec = updatetime;
			tv.tv_usec = 0;
			nfd = select(fvalim + 1, &readfd, &writefd, NULL, &tv);
		}
#ifdef DEBUG
		printf("select() exited with %d descriptors active...\n", nfd);
#endif
		if (nfd < 0) {
			/* an error occurred */
			perror("select()");
			exit(5);
		}

		if (nfd == 0) {
			/* timeout */
			ntimes++;
		}
		
		if (type == SMART_PARALLEL) {
#ifdef DEBUG
			printf("Woken up, now reading values...\n");
#endif
			/* Analog values */
			valim = ADConvert(devaddr, 0);
			temp = ADConvert(devaddr, 1);
			ain1 = ADConvert(devaddr, 2);
			ain2 = ADConvert(devaddr, 3);

			/* digital values (simulated by converting 0V or 5V) */
			status = 0;
			if (ADConvert(devaddr, 4) > 128)
				status = status | 1;
			if (ADConvert(devaddr, 5) > 128)
				status = status | 2;
			if (ADConvert(devaddr, 6) > 128)
				status = status | 4;
			if (ADConvert(devaddr, 7) > 128)
				status = status | 8;
#ifdef DEBUG
			printf("Values read: valim=%2.2f, temp=%2.1f, ain1=%2.2f, ain2=%2.2f, din1=%d, din2=%d, din3=%d, din4=%d\n\n\n", vcc_mult * valim / 255.0, temp_mult * temp / 255.0, ain1_mult * ain1 / 255.0, ain2_mult * ain2 / 255.0, status & 1, status & 2, status & 4, status & 8);
#endif
		}
		/*
		 * See if serial smart interface sent us new data
		 */
		if (type == SMART_SERIAL) {
			if (FD_ISSET(ser, &readfd)) {
				/* data available from interface */
				n += read(ser, buf, sizeof(buf));
				if (n == MSGLEN) {
					/*
					 * Terminate string and parse smart interface message
					 *
					 * Interface sends: Bxx A1xx A2xx Txx Vxx
					 */
					*(buf + n) = '\0';
					n = 0;
#ifdef DEBUG
					printf("Message received from interface: %s\n", buf);
#endif
					r = sscanf(buf, "B%2x A1%2x A2%2x T%2x V%2x", &s, &a1, &a2, &t, &v);

					if (r == 5) {
						status = s;
						ain1 = a1;
						ain2 = a2;
						temp = t;
						valim = v;
#ifdef DEBUG
						printf("Decoded data: status=%2.2x, ain1=%2.2x, ain2=%2.2x, temp=%2.2x, valim=%2.2x\n", status, ain1, ain2, temp, valim);
#endif
					}
				} else if (n > MSGLEN) {
					/* garbage, throw it out */
					n = 0;
					*buf = '\0';
				}
			}
		}

		/*
		 * If we have a smart interface, log every logtime updates analog values
		 */
		if ((type == SMART_SERIAL) || (type == SMART_PARALLEL)) {
		    if (ntimes == logtime) {
			ntimes = 0;
			writelog(logfile, log_analog_values_fmt, vcc_mult * valim / 255, temp_mult * temp / 255,
				 ain1_mult * ain1 / 255, ain2_mult * ain2 / 255);
		    }
		}

		if (type == PARALLEL) {
#ifdef DEBUG
			printf("Woken up, now calling ioctl()...\n");
#endif
			if (ioctl(lp, LPGETSTATUS, &status) != 0) {
				perror("ioctl()");
				exit(5);
			}
#ifdef DEBUG
			printf("lpstatus=%2.2x\n", status);
#endif
		}

		if (type == SERIAL) {
			if (ioctl(ser, TIOCMGET, &status) != 0) {
				perror("ioctl()");
				exit(5);
			}
#ifdef DEBUG
			printf("serstatus=%2.2x\n", status);
#endif
		}

		/*
		 * See if some application request data
		 */
		if (FD_ISSET(fdin1, &writefd)) {
			/* digital input 1 */
			sprintf(buf, "%1d\n", status & 1 ? 1 : 0);
			write(fdin1, buf, strlen(buf));
		}
		if (FD_ISSET(fdin2, &writefd)) {
			/* digital input 2 */
			sprintf(buf, "%1d\n", status & 2 ? 1 : 0);
			write(fdin2, buf, strlen(buf));
		}
		if (FD_ISSET(fdin3, &writefd)) {
			/* digital input 3 */
			sprintf(buf, "%1d\n", status & 4 ? 1 : 0);
			write(fdin3, buf, strlen(buf));
		}
		if (FD_ISSET(fdin4, &writefd)) {
			/* digital input 4 */
			sprintf(buf, "%1d\n", status & 8 ? 1 : 0);
			write(fdin4, buf, strlen(buf));
		}
		if (FD_ISSET(fain1, &writefd)) {
			/* analog input 1 */
			value = ain1_mult * ain1;
			sprintf(buf, "%2.2f\n", value);
			write(fain1, buf, strlen(buf));
		}
		if (FD_ISSET(fain2, &writefd)) {
			/* analog input 2 */
			value = ain2_mult * ain2;
			sprintf(buf, "%2.2f\n", value);
			write(fain2, buf, strlen(buf));
		}
		if (FD_ISSET(ftemp, &writefd)) {
			/* temperature sensor */
			value = temp_mult * temp / 255;
			sprintf(buf, "%2.1f\n", value);
			write(ftemp, buf, strlen(buf));
		}
		if (FD_ISSET(fvalim, &writefd)) {
			/* +13.8 Vcc line */
			value = vcc_mult * valim / 255;
			sprintf(buf, "%2.1f\n", value);
			write(fvalim, buf, strlen(buf));
		}

		/*
		 * Write alarm events into logfile
		 */

		/*
		 * Power events
		 */
		if (power_mask != NONE) {
			if ((status & power_mask) == power_status) {
				if (power_status == power_ok) {
					writelog(logfile, log_event_fmt, power_on_text);
					mailsysop(mailto_power, power_on_text, NULL);
					power_status = power_fail;
				} else {
					writelog(logfile, log_event_fmt, power_off_text);
					mailsysop(mailto_power, power_off_text, NULL);
					power_status = power_ok;
				}
			}
		}
		/*
		 * Write OK / FAIL to stdout to report status to powerd
		 */
		if (power_mask != NONE) {
			if ((status & power_mask) == power_ok) {
				printf("OK\n");
			} else {
				printf("FAIL\n");
			}
		}

		fflush(stdout);
		/*
		 * Generic alarms
		 */
		for (index = 0; index < 4; index++) {
			if ((alarm_mask[index] != NONE) && ((status & alarm_mask[index]) == alarm_status[index])) {
				if (alarm_status[index] == alarm_on[index]) {
					writelog(logfile, log_event_fmt, alarm_on_text[index]);
					mailsysop(mailto_alarm[index], alarm_on_text[index], alarm_on_action[index]);
					if (alarm_on_action[index] != NULL) {
						writelog(logfile, log_action_fmt, alarm_on_action[index]);
						system(alarm_on_action[index]);
					}
					alarm_status[index] = alarm_off[index];
				} else {
					writelog(logfile, log_event_fmt, alarm_off_text[index]);
					mailsysop(mailto_alarm[index], alarm_off_text[index], alarm_on_action[index]);
					if (alarm_off_action[index] != NULL) {
						writelog(logfile, log_action_fmt, alarm_off_action[index]);
						system(alarm_off_action[index]);
					}
					alarm_status[index] = alarm_on[index];
				}
			}
		}
	}
}
